/*
 * Copyright (C) 2006 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, version 2.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 */

#include <ruby.h>
#include <stdbool.h>

#include "vcedit.h"
#include "comments.h"

typedef struct {
	VALUE io;
	bool need_close;

	vcedit_state *state;

	VALUE comments;
} RbVorbisTagger;

static VALUE c_close (VALUE self);

static VALUE cComments, eVTError, io_buf;
static ID id_read, id_write, id_seek, id_length;

static size_t
on_read (void *ptr, size_t size, size_t nmemb, RbVorbisTagger *o)
{
	struct RString *buf;
	size_t total = size * nmemb;
	VALUE tmp;

	rb_str_resize (io_buf, size * nmemb);

	tmp = rb_funcall (o->io, id_read, 2, LONG2NUM (total), io_buf);
	if (NIL_P (tmp))
		return 0;

	buf = RSTRING (tmp);
	memcpy (ptr, buf->ptr, buf->len);

	return buf->len;
}

static size_t
on_write (const void *ptr, size_t size, size_t nmemb, RbVorbisTagger *o)
{
	size_t total = size * nmemb;

	rb_str_resize (io_buf, total);
	memcpy (RSTRING (io_buf)->ptr, ptr, total);

	return NUM2LONG (rb_io_write (o->io, io_buf));
}

static void
c_mark (RbVorbisTagger *o)
{
	rb_gc_mark (o->io);
	rb_gc_mark (o->comments);
}

static void
c_free (RbVorbisTagger *o)
{
	/* just in case the user forgot to call #close himself */
	if (o->state)
		vcedit_state_unref (o->state);

	ruby_xfree (o);
}

static VALUE
c_alloc (VALUE klass)
{
	RbVorbisTagger *o;

	return Data_Make_Struct (klass, RbVorbisTagger, c_mark, c_free, o);
}

/*
 * call-seq:
 *  Ogg::Vorbis::Tagger.open(arg)                    -> object
 *  Ogg::Vorbis::Tagger.open(arg) { |object| block } -> nil
 *
 * If a block isn't specified, Ogg::Vorbis::Tagger.open is a synonym
 * for Ogg::Vorbis::Tagger.new.
 * If a block is given, it will be invoked with the
 * Ogg::Vorbis::Tagger object as a parameter, and the file or IO object
 * will be automatically closed when the block terminates.
 * The method always returns +nil+ in this case.
 */
static VALUE
c_open (VALUE klass, VALUE arg)
{
	VALUE obj = rb_class_new_instance (1, &arg, klass);

	if (rb_block_given_p ())
		return rb_ensure (rb_yield, obj, c_close, obj);
	else
		return obj;
}

/*
 * call-seq:
 *  Ogg::Vorbis::Tagger.new(arg) -> object
 *
 * Returns a new Ogg::Vorbis::Tagger object for the specified argument.
 * *arg* can either be an IO object or a filename.
 *
 * FIXME: add optional mode argument (read-only or read-write)
 */
static VALUE
c_init (VALUE self, VALUE io)
{
	RbVorbisTagger *o;
	vorbis_comment *vc;
	int s, i;

	Data_Get_Struct (self, RbVorbisTagger, o);

	/* is this actually an IO object or a filename? */
	if (rb_respond_to (io, id_read) &&
	    rb_respond_to (io, id_write) &&
	    rb_respond_to (io, id_seek))
		o->need_close = false;
	else if (!NIL_P (rb_check_string_type (io))) {
		io = rb_file_open (StringValuePtr (io), "rb+");
		o->need_close = true;
	} else
		rb_raise (rb_eArgError, "invalid argument");

	o->io = io;

	o->state = vcedit_state_new ();
	if (!o->state)
		rb_raise (eVTError, "vcedit_new_state() failed - %s",
		          vcedit_error (o->state));

	s = vcedit_open_callbacks (o->state, o,
	                           (vcedit_read_func) on_read,
	                           (vcedit_write_func) on_write);
	if (s < 0)
		rb_raise (eVTError, "vcedit_open_callbacks() failed - %s",
		          vcedit_error (o->state));

	vc = vcedit_comments (o->state);
	if (!vc)
		rb_raise (eVTError, "vcedit_comments() failed - %s",
		          vcedit_error (o->state));

	/* check whether all comments are well-formed */
	for (i = 0; i < vc->comments; i++) {
		char *ptr, *content = vc->user_comments[i];

		ptr = strchr (content, '=');
		if (!ptr || ptr == content)
			rb_raise (eVTError, "malformed comment - %s", content);
	}

	o->comments = rb_class_new_instance (0, NULL, cComments);

	comments_init (o->comments, o->state);

	return self;
}

/*
 * call-seq:
 *  object.close -> object
 *
 * Closes *object* and returns it.
 */
static VALUE
c_close (VALUE self)
{
	RbVorbisTagger *o;

	Data_Get_Struct (self, RbVorbisTagger, o);

	vcedit_state_unref (o->state);
	o->state = NULL;

	if (o->need_close)
		rb_io_close (o->io);

	return self;
}

/*
 * call-seq:
 *  object.write -> integer
 *
 * Writes the comments from *object* back to the IO object and
 * returns the numbers of comments written.
 */
static VALUE
c_write (VALUE self)
{
	RbVorbisTagger *o;
	int s;

	Data_Get_Struct (self, RbVorbisTagger, o);

	comments_sync (o->comments, o->state);

	/* seek back to BOF */
	rb_funcall (o->io, id_seek, 1, INT2FIX (0));

	s = vcedit_write (o->state, o);
	if (s < 0)
		rb_raise (rb_eIOError, "write failed - %s",
		          vcedit_error (o->state));

	return rb_funcall (o->comments, id_length, 0);
}

/*
 * call-seq:
 *  object.comments -> comments
 *
 * Returns the comments collection of *object*, which is an instance of
 * Ogg::Vorbis::Comments.
 */
static VALUE
c_comments (VALUE self)
{
	RbVorbisTagger *o;

	Data_Get_Struct (self, RbVorbisTagger, o);

	return o->comments;
}

EXT_API
void
Init_vorbistagger_ext (void)
{
	VALUE mOgg, mVorbis, eOgg, eVorbis, cVT;

	mOgg = rb_define_module ("Ogg");
	mVorbis = rb_define_module_under (mOgg, "Vorbis");

	eOgg = rb_define_class_under (mOgg, "OggError", rb_eStandardError);
	eVorbis = rb_define_class_under (mVorbis, "VorbisError", eOgg);

	cVT = rb_define_class_under (mVorbis, "Tagger", rb_cObject);

	rb_define_alloc_func (cVT, c_alloc);

	rb_define_singleton_method (cVT, "open", c_open, 1);
	rb_define_method (cVT, "initialize", c_init, 1);
	rb_define_method (cVT, "close", c_close, 0);
	rb_define_method (cVT, "write", c_write, 0);
	rb_define_method (cVT, "comments", c_comments, 0);

	eVTError = rb_define_class_under (cVT, "TaggerError", eVorbis);

	id_read = rb_intern ("read");
	id_write = rb_intern ("write");
	id_seek = rb_intern ("seek");
	id_length = rb_intern ("length");

	cComments = Init_Comments (mVorbis);

	io_buf = rb_str_buf_new (0);
	rb_global_variable (&io_buf);
}
